/*******************************************************************************
 * Copyright (c) 2008 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/

package org.eclipse.birt.report.engine.api.impl;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.birt.report.engine.api.IReportDocument;
import org.eclipse.birt.report.engine.api.ITOCTree;
import org.eclipse.birt.report.engine.api.TOCNode;
import org.eclipse.birt.report.engine.api.script.instance.IScriptStyle;

public class TOCView implements ITOCTree
{

	IReportDocument document;
	ITOCTree tocTree;
	TOCViewNode viewRoot;
	LogicalPageSequence visiblePages;

	TOCView( ITOCTree tocTree, IReportDocument document,
			LogicalPageSequence visiblePages )
	{
		this.document = document;
		this.tocTree = tocTree;
		this.visiblePages = visiblePages;
		this.viewRoot = createViewNode( null, tocTree.getRoot( ) );
	}

	public TOCNode findTOC( String tocId )
	{
		if ( tocId == null || "/".equals( tocId ) )
		{
			return viewRoot;
		}

		TOCNode node = tocTree.findTOC( tocId );
		if ( isVisible( node ) )
		{
			return createViewNode( null, node );
		}
		return null;
	}

	public List findTOCByValue( Object tocValue )
	{
		List nodeList = tocTree.findTOCByValue( tocValue );
		ArrayList viewList = new ArrayList( );
		Iterator iter = nodeList.iterator( );
		while ( iter.hasNext( ) )
		{
			TOCNode node = (TOCNode) iter.next( );
			if ( isVisible( node ) )
			{
				viewList.add( createViewNode( null, node ) );
			}

		}
		return viewList;
	}

	public TOCNode getRoot( )
	{
		return viewRoot;
	}

	class TOCViewNode extends TOCNode
	{

		TOCNode node;
		TOCViewNode viewParent;
		ArrayList viewChildren;

		TOCViewNode( TOCViewNode parent, TOCNode node )
		{
			this.viewParent = parent;
			this.node = node;
			this.viewChildren = new ArrayList( );
		}

		public String getNodeID( )
		{
			return node.getNodeID( );
		}

		public TOCNode getParent( )
		{
			return viewParent;
		}

		public String getDisplayString( )
		{
			return node.getDisplayString( );
		}

		public String getBookmark( )
		{
			return node.getBookmark( );
		}

		public IScriptStyle getTOCStyle( )
		{
			return node.getTOCStyle( );
		}

		public List getChildren( )
		{
			return viewChildren;
		}
	}

	boolean isVisible( TOCNode node )
	{
		String id = node.getNodeID( );
		if ( id == null || "/".equals( node ) )
		{
			return true;
		}
		String bookmark = node.getBookmark( );
		long physicalPageNumber = document.getPageNumber( bookmark );
		long pageNumber = visiblePages
				.getLogicalPageNumber( physicalPageNumber );

		return pageNumber != -1;
	}

	TOCViewNode createViewNode( TOCViewNode parent, TOCNode node )
	{
		TOCViewNode viewNode = null;
		if ( isVisible( node ) )
		{
			viewNode = new TOCViewNode( parent, node );
		}

		List children = node.getChildren( );
		Iterator iter = children.iterator( );
		if ( viewNode == null )
		{
			while ( iter.hasNext( ) )
			{
				TOCNode child = (TOCNode) iter.next( );
				viewNode = createViewNode( parent, child );
				if ( viewNode != null )
				{
					break;
				}
			}
		}

		if ( viewNode != null )
		{
			List viewChildren = viewNode.getChildren( );
			while ( iter.hasNext( ) )
			{
				TOCNode child = (TOCNode) iter.next( );
				TOCViewNode viewChild = createViewNode( viewNode, child );
				if ( viewChild != null )
				{
					viewChildren.add( viewChild );
				}
			}
		}

		return viewNode;
	}
}
